import math

import numpy as np

from shor.layers import _BaseLayer


class _Gate(_BaseLayer):
    """Abstract base quantum gate class

    # Properties
    input_length = valid length of input qubits
    qubits = indices of qubits, to be used as input to gate.
    """

    def __init__(self, *qubits: int, **kwargs):
        super().__init__(**kwargs)
        self.dimension = kwargs.get('dimension', 1)
        self.qubits = qubits if qubits else [0]

        assert all(map(lambda q: type(q) == int, self.qubits))
        assert len(self.qubits) == self.dimension

    def to_gates(self):
        return [self]

    @property
    def num_states(self):
        return np.power(2, self.dimension)


class CNOT(_Gate):
    def __init__(self, *qubits, **kwargs):
        kwargs['dimension'] = 2
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])


class CSWAP(_Gate):
    def __init__(self, *qubits, **kwargs):
        kwargs['dimension'] = 3
        if not qubits:
            qubits = [0, 1, 2]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        cswap_matrix = np.eye(8)
        cswap_matrix[:, [5, 6]] = cswap_matrix[:, [6, 5]]
        return cswap_matrix


class Hadamard(_Gate):
    def __init__(self, *qubits, **kwargs):
        kwargs['dimension'] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.multiply(np.divide(1, np.sqrt(self.num_states)), np.array([[1, 1], [1, -1]]))


class PauliX(_Gate):
    def __init__(self, *qubits, **kwargs):
        kwargs['dimension'] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[0, 1], [1, 0]])


class PauliY(_Gate):
    def __init__(self, *qubits, **kwargs):
        kwargs['dimension'] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[0, -1j], [1j, 0]])


class PauliZ(_Gate):
    def __init__(self, *qubits, **kwargs):
        kwargs['dimension'] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0], [0, -1]])


class QFT(_Gate):
    def __init__(self, *qubits, **kwargs):
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, dimension=len(qubits), **kwargs)

    # def to_gates(self):
    #     # TODO: translate this gate to base gates / CNOTs
    #     pass

    def get_nth_unity_root(self, k):
        return np.exp((2j * np.pi * k) / self.num_states)

    def to_matrix(self) -> np.ndarray:
        m = np.array(np.ones((self.num_states, self.num_states)), dtype='complex')

        for i in range(1, self.num_states):
            for j in range(i, self.num_states):
                w = self.get_nth_unity_root(i * j)
                m[i, j] = w
                m[j, i] = w

        return np.around(
            np.multiply(
                1 / np.sqrt(self.num_states),
                m
            ), decimals=15)


class Rx(_Gate):
    def __init__(self, *qubits, angle=math.pi/2, **kwargs):
        kwargs['dimension'] = 1
        self.angle=angle
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.array([
        [math.cos(self.angle / 2), -math.sin(self.angle / 2) * 1j],
        [-math.sin(self.angle / 2) * 1j, math.cos(self.angle / 2)]
    ])


class SWAP(_Gate):
    def __init__(self, *qubits, **kwargs):
        kwargs['dimension'] = 2
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])


# Aliases
H = h = Hadamard
X = x = PauliX
Y = y = PauliY
Z = z = PauliZ
swap = SWAP
CX = cx = CNOT
Fredkin = cswap = CSWAP



import math

import numpy as np

from tests.util import is_square, is_unitary


def test_cnot_init():
    from shor.gates import CNOT
    CNOT()


def test_hadamard_init():
    from shor.gates import Hadamard, H
    gate1 = H()
    gate2 = Hadamard()

    assert gate1.__class__ == gate2.__class__

    # Try with parameter
    H(0)


def test_paulix_init():
    from shor.gates import PauliX, X
    gate1 = X()
    gate2 = PauliX()
    assert gate1.__class__ == gate2.__class__
    # Try with parameter
    X(0)


def test_pauliy_init():
    from shor.gates import PauliY, Y
    gate1 = Y()
    gate2 = PauliY()
    assert gate1.__class__ == gate2.__class__
    # Try with parameter
    Y(0)


def test_pauliz_init():
    from shor.gates import PauliZ, Z
    gate1 = Z()
    gate2 = PauliZ()
    assert gate1.__class__ == gate2.__class__
    # Try with parameter
    Z(0)


def test_qft_init():
    from shor.gates import QFT
    g = QFT(0, 1)


def test_swap_init():
    from shor.gates import SWAP
    g = SWAP()

    # Try with parameters
    SWAP(1, 0)


def test_rx():
    from shor.gates import Rx
    angle = math.pi / 8
    g = Rx(0, angle=math.pi / 8)

    assert is_square(g.to_matrix())
    assert is_unitary(g.to_matrix())

    assert np.array_equal(g.to_matrix(), np.array([
        [math.cos(angle / 2), -math.sin(angle / 2) * 1j],
        [-math.sin(angle / 2) * 1j, math.cos(angle / 2)]
    ]))


def test_cnot_matrix():
    from shor.gates import CNOT
    g = CNOT()

    assert is_square(g.to_matrix())
    assert is_unitary(g.to_matrix())
    assert np.array_equal(g.to_matrix(), np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]))

    # Try with parameter
    CNOT(1, 0)


def test_cswap_matrix():
    from shor.gates import CSWAP, Fredkin
    gate1 = CSWAP()
    gate2 = Fredkin()

    assert gate1.__class__ == gate2.__class__

    assert is_square(gate1.to_matrix())
    assert is_unitary(gate1.to_matrix())

    expected = np.eye(8)
    expected[:, [5, 6]] = expected[:, [6, 5]]
    assert np.array_equal(gate1.to_matrix(), expected)

    # Try with parameter
    CSWAP(2, 0, 1)


def test_hadamard_matrix():
    from shor.gates import Hadamard, H
    gates = [Hadamard(), H()]

    for g in gates:
        assert is_square(g.to_matrix())
        assert is_unitary(g.to_matrix())
        assert np.array_equal(g.to_matrix(), np.multiply(np.divide(1, np.sqrt(2)), np.array([[1, 1], [1, -1]])))


def test_paulix_matrix():
    from shor.gates import PauliX, X
    gates = [PauliX(), X()]

    for g in gates:
        assert is_square(g.to_matrix())
        assert is_unitary(g.to_matrix())
        assert np.array_equal(g.to_matrix(), np.array([[0, 1], [1, 0]]))


def test_qft_matrix():
    from shor.gates import QFT

    g = QFT(0, 1)

    assert is_square(g.to_matrix())
    assert is_unitary(g.to_matrix())
    assert np.array_equal(g.to_matrix(), np.multiply(
        1/2,
       np.array([[1, 1, 1, 1], [1, 1j, -1, -1j], [1, -1, 1, -1], [1, -1j, -1, 1j]])
    ))


def test_swap_matrix():
    from shor.gates import SWAP
    g = SWAP(0, 1)

    assert is_square(g.to_matrix())
    assert is_unitary(g.to_matrix())
    assert np.array_equal(g.to_matrix(), np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]))



from shor.gates import CNOT, Hadamard
from shor.layers import Qubits
from shor.operations import Measure
from shor.quantum import Circuit
from shor.backends import QuantumSimulator, QSession


def test_single_qubit():
    circuit = Circuit()
    circuit.add(Qubits(1))
    circuit.add(Hadamard())  # Can also use H()
    circuit.add(Measure())

    from shor.backends import QuantumSimulator, QSession

    sess = QSession(backend=QuantumSimulator())
    result = sess.run(circuit, num_shots=1024)

    # Accounting for random noise, results won't be exact
    assert result[bin(0)] > 450
    assert result[bin(1)] > 450


def test_entanglement():
    circuit = Circuit()
    circuit.add(Qubits(2))
    circuit.add(Hadamard(0))
    circuit.add(CNOT(0, 1))
    circuit.add(Measure(0, 1))

    sess = QSession(backend=QuantumSimulator())
    result = sess.run(circuit, num_shots=1024)

    assert result['01'] == 0
    assert result['10'] == 0
    assert result['00'] > 450
    assert result['11'] > 450


def test_unitary_symmetry_does_nothing():
    symmetric_circuit_1 = Circuit()
    symmetric_circuit_1.add(Qubits(2))
    symmetric_circuit_1.add(Hadamard(0))
    symmetric_circuit_1.add(Hadamard(0))
    symmetric_circuit_1.add(CNOT(0, 1))
    symmetric_circuit_1.add(CNOT(0, 1))
    symmetric_circuit_1.add(Measure(0, 1))

    symmetric_circuit_2 = Circuit()
    symmetric_circuit_2.add(Qubits(2, state=1))
    symmetric_circuit_2.add(Hadamard(0))
    symmetric_circuit_2.add(CNOT(0, 1))
    symmetric_circuit_2.add(CNOT(0, 1))
    symmetric_circuit_2.add(Hadamard(0))
    symmetric_circuit_2.add(Measure(0, 1))

    sess = QSession(backend=QuantumSimulator())
    result_1 = sess.run(symmetric_circuit_1, num_shots=1024)
    result_2 = sess.run(symmetric_circuit_2, num_shots=1024)

    assert result_1.counts.get(0) == 1024
    assert result_2.counts.get(3) == 1024


def test_multi_entangle():
    circuit = Circuit()
    circuit.add(Qubits(4))
    circuit.add(Hadamard(0))
    circuit.add(CNOT(0, 1))
    circuit.add(CNOT(0, 2))
    circuit.add(CNOT(0, 3))
    circuit.add(Measure(0, 1, 2, 3))

    sess = QSession(backend=QuantumSimulator())
    result = sess.run(circuit, num_shots=1024)

    assert result['0001'] == 0
    assert result['1000'] == 0
    assert result['0000'] > 450
    assert result['1111'] > 450


def test_multi_hadamard():
    circuit = Circuit()
    circuit.add(Qubits(4))
    circuit.add(Hadamard(0))
    circuit.add(Hadamard(1))
    circuit.add(Hadamard(2))
    circuit.add(Hadamard(3))
    circuit.add(Measure(0, 1, 2, 3))

    sess = QSession(backend=QuantumSimulator())
    result = sess.run(circuit, num_shots=1024)

    # All 16 states should be relatively equal probability
    assert len(result.counts) == 16
    assert max(result.counts.values()) - min(result.counts.values()) < 50




from shor.gates import H
from shor.layers import Qbits


def test_quantum_backend_init():
    from shor.backends import _QuantumBackend
    _QuantumBackend()


def test_quantum_simulator_init():
    from shor.backends import QuantumSimulator
    QuantumSimulator()


def test_qsession_init():
    from shor.backends import QSession
    QSession()


def test_qsession_run():
    from shor.backends import QSession
    sess = QSession()

    from shor.quantum import Circuit
    sess.run(Circuit().add(Qbits(1)).add(H()), num_shots=10)


def test_qresult_str_representation():
    from shor.backends import QResult

    result = QResult({0: 20, 7: 34}, 5)

    assert str(result) == repr(result) == "{'00000': 20, '00111': 34}"


def test_qresult_get():
    from shor.backends import QResult

    result = QResult({0: 20, 7: 34}, 5)

    assert 20 == result['00000'] == result['0'] == result[0] == result[0b00000] == result.get(0)
    assert 34 == result['00111'] == result['111'] == result[7] == result[0b00111] == result[0b111] == result.get(7)



from shor.layers import _BaseLayer


class _Operation(_BaseLayer):
    """Abstract base quantum computing operation class
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def to_gates(self):
        return []


class Measure(_Operation):
    def __init__(self, *bits, axis='z', **kwargs):
        if not bits:
            bits = [0]
        self.bits = bits

        super().__init__(**kwargs, axis=axis)



class _BaseLayer(object):
    """Abstract base quantum layer class"""

    def __init__(self, **kwargs):
        pass

    def to_gates(self):
        pass


class _Layer(_BaseLayer):
    """Abstract base quantum layer class

    # Properties
    inputs = indices of qubits, to be used as inputs.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def to_gates(self):
        pass


class Qubits(_BaseLayer):
    def __init__(self, num, state=0, **kwargs):
        self.num = num
        self.state = state

        super().__init__(**kwargs)

    def to_gates(self):
        return []


# Aliasing class
Qbits = Qubits



import pytest

from shor.quantum import Circuit
from shor.layers import _BaseLayer


def test_circuit_init():
    Circuit()


def test_circuit_add_base_layer():
    circuit = Circuit()
    prev_num_layers = len(circuit.layers)
    circuit.add(_BaseLayer())

    assert len(circuit.layers) == prev_num_layers + 1


def test_circuit_add_returns_self():

    circuit1 = Circuit()
    circuit2 = circuit1.add(_BaseLayer())

    assert circuit1 == circuit2


def test_circuit_add_circuit():
    circuit = Circuit().add(_BaseLayer()).add(_BaseLayer())

    circuit_to_add = Circuit()
    circuit_to_add.add(_BaseLayer()).add(_BaseLayer()).add(_BaseLayer())

    assert len(circuit.layers) == 2
    circuit.add(circuit_to_add)

    assert len(circuit.layers) == 5


def test_circuit_add_wrong_type():
    class SomeClass(object):
        def __init__(self):
            pass

    with pytest.raises(TypeError) as type_error:
        circuit = Circuit().add(SomeClass())